/* face.c
 * Nadim Abdo 1998
 *
 * Implements a face process
 *
 * For Operating Systems 304-427A Assignment #3
 */

/* =================================================== */
/* ==================== Includes ===================== */
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/ipc.h>
#include <sys/shm.h>  
#include <sys/sem.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/msg.h>

#include "semaph.h" /*Steven's Unix Net Programming sem utils */
#include "life3d.h"
#include "face.h"



/*Copy neighbouring edges into a local [N+2][N+2] array*/
/*Should only call this if we have a lock on the shared mem*/
void CopyEdgeToLocalArray(int local_startx,int local_starty,
                          walk_dir local_walk_direction,
                          int facenum,
                          int edge,
                          AxisDir dir,
                          int facearray[][N+2])
{
  int remote_startx,remote_starty;
  walk_dir remote_walkdir;
  int i,j;
  int cell_state;
  
  remote_startx  = LUTCopyOrigin[edge].startx;
  remote_starty  = LUTCopyOrigin[edge].starty;
  remote_walkdir = LUTCopyOrigin[edge].dir;

  for(i=0;i<N;i++)
    {
      /*Get the next cell from the destination*/
      
      switch (remote_walkdir)
        {
        case DirHorizontal:
          cell_state=pLifeState->cells[facenum][remote_startx+i][remote_starty];
          break;
        case DirVertical:
          cell_state=pLifeState->cells[facenum][remote_startx][remote_starty+i];
          break;
        }

      /*Now assign it to the local face*/
      /*Walk backwards if the axes are in opposing directions*/
      j = (RevDir == dir) ? N-1-i : i;
      switch (local_walk_direction)
        {
        case DirHorizontal:
          facearray[local_startx+j][local_starty]=cell_state;
          break;
        case DirVertical:
          facearray[local_startx][local_starty+j]=cell_state;
          break;
        }
    }
}


/*Function assumes we have a lock on the shared memory*/
void SetupLocalFaces(int facenum, int facearray[][N+2])
{
  int i,j;

  /*First be extra careful and kill everything
    in the local face map*/

  for(i=0;i<(N+2);i++)
      for(j=0;j<(N+2);j++)
          facearray[i][j]=CELL_DEAD;
  
  /*Copy in the center portion from shared mem*/
  for(i=0;i<N;i++)
      for(j=0;j<N;j++)
          facearray[i+1][j+1]=pLifeState->cells[facenum][i][j];

  
  /*Now copy in the neighbouring edges*/
  /*North edge*/
  CopyEdgeToLocalArray(1,0,DirHorizontal,
                       face_neighbours[facenum].nbors[NORTH].face,
                       face_neighbours[facenum].nbors[NORTH].edge,
                       face_neighbours[facenum].nbors[NORTH].dir,
                       facearray);

  /*West*/
  CopyEdgeToLocalArray(0,1,DirVertical,
                       face_neighbours[facenum].nbors[WEST].face,
                       face_neighbours[facenum].nbors[WEST].edge,
                       face_neighbours[facenum].nbors[WEST].dir,
                       facearray);
  /*South*/
  CopyEdgeToLocalArray(1,N+1,DirHorizontal,
                       face_neighbours[facenum].nbors[SOUTH].face,
                       face_neighbours[facenum].nbors[SOUTH].edge,
                       face_neighbours[facenum].nbors[SOUTH].dir,
                       facearray);
  /*East*/
  CopyEdgeToLocalArray(N+1,1,DirVertical,
                       face_neighbours[facenum].nbors[EAST].face,
                       face_neighbours[facenum].nbors[EAST].edge,
                       face_neighbours[facenum].nbors[EAST].dir,
                       facearray);

  
}


/*Compute the next state of the cells in the inner NxN cells
  of the given [N+2][N+2] array.
  This is nice because a) the neigbour cells are automatically
  accounted for.
  There is no need to check array bounds.
*/
void ComputeLifeNextState(int facearray[][N+2])
{
  int i,j;
  int n;
  
  for(i=1;i<(N+1);i++)
    {
      for(j=1;j<(N+1);j++)
        {
          n=0;
          if((CELL_ALIVE & facearray[i-1][j]))
            n++;
          if((CELL_ALIVE & facearray[i+1][j]))
            n++;
          if((CELL_ALIVE & facearray[i][j-1]))
            n++;
          if((CELL_ALIVE & facearray[i][j+1]))
            n++;
          /*diagonals*/
          if((CELL_ALIVE & facearray[i+1][j+1]))
            n++;
          if((CELL_ALIVE & facearray[i-1][j-1]))
            n++;
          if((CELL_ALIVE & facearray[i+1][j-1]))
            n++;
          if((CELL_ALIVE & facearray[i-1][j+1]))
            n++;
          
          
          /*A cell is alive on the next generation
            if
            a) it is dead and has 3 neighbours
            b) it is alive and has 2 or 3 neighbours
          */
          if( (3==n)  ||
              ((2==n) && (CELL_ALIVE & facearray[i][j])))
            {
              facearray[i][j]|=CELL_ALIVE_NEXT;
            }
          
          else
            facearray[i][j]&=CELL_DEAD_NEXT;
        }
    }
}

void UpdateCubeWithNextState(int facearray[][N+2],int facenum)
{
  int i,j;
  
  for(i=0;i<N;i++)
    for(j=0;j<N;j++)
      {
        if(facearray[i+1][j+1] & CELL_ALIVE_NEXT)
          pLifeState->cells[facenum][i][j]=CELL_ALIVE;
        else
          pLifeState->cells[facenum][i][j]=CELL_DEAD;
      }
}


/*Runs life on a given face*/
void RunLifeOnFace(int facenum)
{
  /*The game of life runs forever*/
  /*temporary array large enough for a face + its neighbours*/
  int localface[N+2][N+2];
  
  
  for(;;)
    {
      /*wait for the start of a time slice*/
      semwait(sem_time_slice_start);

      /*Get a lock on the shared mem*/
      semwait(sem_shared_mem);
      SetupLocalFaces(facenum,localface);
      semsignal(sem_shared_mem);


      /*Compute the next state of the local cells*/
      ComputeLifeNextState(localface);
#ifdef WANNA_DEBUG_ME
      printf("FACE%d************\n",facenum+1);
      for(i=0;i<(N+2);i++)
        {
          
          for(j=0;j<(N+2);j++)
            printf("%d ",localface[j][i]);
          printf("\n");
        }
#endif
      
      /*Signal that this face is done updating it's next state
        the main process waits for all the faces to reach this
        point before commiting the next state to the current state*/
      semsignal(sem_num_updated_faces);

      /*Synchronize the face's end of time slice to the main process*/
      semwait(sem_time_slice_end);

      semwait(sem_shared_mem);
      UpdateCubeWithNextState(localface,facenum);
      semsignal(sem_shared_mem);
      
      /*Indicate this face has commited next state to current state*/
      semsignal(sem_faces_commited_next_state);
      
    }
}
